# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
# Copyright (c) 2019, Eurecat / UPF
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the <organization> nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
#   @file   apply_source_signals.py
#   @author Andrés Pérez-López
#   @date   30/07/2019
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

import numpy as np
import scipy.signal

from masp.validate_data_types import _validate_ndarray_3D, _validate_ndarray_2D, _validate_ndarray_4D, _validate_list


def apply_source_signals_array(array_rirs, src_sigs):
    """
    Apply room impulse responses from an array of receivers (microphone arrays) to a set of source signals.

    Parameters
    ----------
    array_rirs : List
      RIR for each receiver element. Length = (nRec)
    src_sigs: ndarray
       Matrix containing the source signals. Dimension = (L_sig, nSrc)

    Returns
    -------
    array_sigs : List
        Source signals subjected to the RIRs. Length = (nRec)

    Raises
    -----
    TypeError, ValueError: if method arguments mismatch in type, dimension or value.

    Notes
    -----
    Each element in `array_rirs` should be a ndarray wit dimension = (L, nMic, nSrc).
    All values of `nSrc` across receivers should be equal, and also match (shape[1]) in `src_sigs`.

    Each element of the algorithm output `array_sigs` contains the rendering of the sources from a different receiver,
    featuring a ndarray with dimension = (L_rir+L_sig-1, nMic).

    TODO: check return values
    """

    _validate_list('array_rirs', array_rirs)
    nRec = len(array_rirs)
    nSrcs = [array_rirs[i].shape[2] for i in range(nRec)]
    assert nSrcs[1:] == nSrcs[:-1]  # Same number of sources for each receiver
    nSrc = nSrcs[0]
    _validate_ndarray_2D('src_sigs', src_sigs, shape1=nSrc)

    array_sigs = [None] * nRec      # Empty list
    src_array_sigs = [None] * nRec  # Empty list
    L_sig = src_sigs.shape[0]

    for nr in range(nRec):
        temprirs = array_rirs[nr]
        L_rir = temprirs.shape[0]
        nMics = temprirs.shape[1]
        tempsigs = np.zeros((L_sig + L_rir - 1, nMics, nSrc))
        for ns in range(nSrc):
            print('Convolving with source signal: Source ' + str(ns) + ' - Receiver ' + str(nr))
            tempsigs[:, :, ns] = scipy.signal.fftconvolve(temprirs[:, :, ns], src_sigs[:, ns, np.newaxis], axes=0)
        src_array_sigs[nr] = tempsigs
        array_sigs[nr] = np.sum(tempsigs, axis=2)

    # return array_sigs, src_array_sigs
    return array_sigs


def apply_source_signals_mic(mic_rirs, src_sigs):
    """
    Apply microphone room impulse responses to a set of source signals.

    Parameters
    ----------
    mic_rirs : ndarray
       Matrix containing the room impulse responses. Dimension = (L_rir, nRec, nSrc)
    src_sigs: ndarray
       Matrix containing the source signals. Dimension = (L_sig, nSrc)

    Returns
    -------
    mic_sigs : ndarray
        Source signals subjected to the RIRs. Dimension = = (L_rir+L_sig-1, nRec)

    Raises
    -----
    TypeError, ValueError: if method arguments mismatch in type, dimension or value.

    Notes
    -----
    The number of source positions (shape[2]) in `mic_rirs` should match the number of sources (shape[1]) in `src_sigs`.

    """
    L_rir = mic_rirs.shape[0]
    nRec = mic_rirs.shape[1]
    nSrc = mic_rirs.shape[2]
    L_sig = src_sigs.shape[0]
    _validate_ndarray_3D('mic_rirs', mic_rirs)
    _validate_ndarray_2D('src_sigs', src_sigs, shape1=nSrc)

    mic_sigs = np.zeros((L_sig + L_rir - 1, nRec))
    for nr in range(nRec):
        for ns in range(nSrc):
            print('Convolving with source signal: Source ' + str(ns) + ' - Receiver ' + str(nr))
            mic_sigs[:, nr] = mic_sigs[:, nr] + scipy.signal.fftconvolve(mic_rirs[:, nr, ns], src_sigs[:, ns])

    return mic_sigs


def apply_source_signals_sh(sh_rirs, src_sigs):
    """
    Apply spherical harmonic room impulse responses to a set of source signals.

    Parameters
    ----------
    sh_rirs : ndarray
       Matrix containing the room impulse responses. Dimension = (L_rir, nSH, nRec, nSrc)
    src_sigs: ndarray
       Matrix containing the source signals. Dimension = (L_sig, nSrc)

    Returns
    -------
    sh_sigs : ndarray
        Source signals subjected to the RIRs. Dimension = = (L_rir+L_sig-1, nSH, nRec)

    Raises
    -----
    TypeError, ValueError: if method arguments mismatch in type, dimension or value.

    Notes
    -----
    The number of source positions (shape[3]) in `sh_rirs` should match the number of sources (shape[1]) in `src_sigs`.

    TODO: check return values
    """

    L_rir = sh_rirs.shape[0]
    nSH = sh_rirs.shape[1]
    nRec = sh_rirs.shape[2]
    nSrc = sh_rirs.shape[3]
    L_sig = src_sigs.shape[0]
    _validate_ndarray_4D('mic_rirs', sh_rirs)
    _validate_ndarray_2D('src_sigs', src_sigs, shape1=nSrc)

    sh_sigs = np.zeros((L_sig + L_rir - 1, nSH, nRec))
    src_sh_sigs = np.zeros((L_sig + L_rir - 1, nSH, nRec, nSrc))
    for nr in range(nRec):
        for ns in range(nSrc):
            print('Convolving with source signal: Source ' + str(ns) + ' - Receiver ' + str(nr))
            # Just apply convolution to non-zero channels (skip empty channels due to different sh orders)
            idx_nonzero = np.sum(sh_rirs[:,:, nr, ns],axis=0) != 0
            src_sh_sigs[:, idx_nonzero, nr, ns] = scipy.signal.fftconvolve(sh_rirs[:, idx_nonzero, nr, ns], src_sigs[:, ns, np.newaxis], axes=0)
    sh_sigs = np.sum(src_sh_sigs, axis=3)

    # return sh_sigs, src_sh_sigs
    return sh_sigs




import setuptools
import masp.version

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="masp",
    version=masp.version.get_API_version(),
    author="Andres Perez-Lopez",
    author_email="contact@andresperezlopez.com",
    url="https://github.com/andresperezlopez/masp",
    description="Multichannel Acoustic Signal Processing",
    long_description="MASP: a Multichannel Acoustic Signal Processing library for Python",
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
# Copyright (c) 2019, Eurecat / UPF
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the <organization> nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
#   @file   validate_data_types.py
#   @author Andrés Pérez-López
#   @date   31/07/2019
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

import numpy as np
from masp.shoebox_room_sim.echogram import Echogram, QuantisedEchogram

def _validate_boolean(name, boolean):

    if not isinstance(boolean, bool):
        raise TypeError(name + ' must be an instance of bool')

def _validate_int(name, number, positive=False, limit=None, parity=None):

    if not isinstance(number, int) or isinstance(number, bool):
        raise TypeError(name + ' must be an instance of int')
    if positive:
        if number < 0:
            raise ValueError(name + ' must be positive')
    if limit is not None:
        if number > limit:
            raise ValueError(name + ' must be smaller than ' + str(limit))
    if parity is not None:
        if parity == 'even':
            if number%2 != 0:
                raise ValueError(name + ' must be even')
        elif parity == 'odd':
            if number%2 == 0:
                raise ValueError(name + ' must be odd')
        else:
            raise TypeError(name + ': unknown parity value: ' + parity)

def _validate_float(name, number, positive=False, limit=None):

    if not isinstance(number, float):
        raise TypeError(name + ' must be an instance of float')
    if positive:
        if number < 0:
            raise ValueError(name + ' must be positive')
    if limit is not None:
        if number > limit:
            raise ValueError(name + ' must be smaller than ' + str(limit))

def _validate_number(name, number, norm=False, positive=False, limit=None):

    if isinstance(number, (int, float)) and not isinstance(number, bool):
        if norm:
            if not 0 <= number <= 1:
                raise ValueError(name + ' must be in the interval [0,1]')
        if positive:
            if 0 > number:
                raise ValueError(name + ' must be positive')
        if limit is not None:
            if number < limit[0]:
                raise ValueError(name + ' must be greater or equal than ' + str(limit[0]))
            if number > limit[1]:
                raise ValueError(name + ' must be smaller or equal than ' + str(limit[1]))
    elif isinstance(number, np.ndarray):
        if norm:
            if np.any(number < 0) or np.any(number > 1):
                raise ValueError(name + ' must be in the interval [0,1]')
        if positive:
            if np.any(number < 0):
                raise ValueError(name + ' must be positive')
        if limit is not None:
            if np.any(number < limit[0]):
                raise ValueError(name + ' must be greater or equal than ' + str(limit[0]))
            if np.any(number > limit[1]):
                raise ValueError(name + ' must be smaller or equal than ' + str(limit[1]))
    else:
        raise TypeError(name + ' must be an integer, float or 1-D ndarray')

def _validate_list(name, arg, size=None):

    if not isinstance(arg, list):
        raise TypeError(name + ' must be an instance of list')
    if size is not None:
        if len(arg) != size:
            raise ValueError(name + ' must have size ' + str(size))

def _validate_ndarray(name, ndarray):

    if not isinstance(ndarray, np.ndarray):
        raise TypeError(name + ' must be an instance of ndarray')

def _validate_ndarray_1D(name, ndarray, size=None, norm=False, positive=False, limit=None, dtype=None):

    if not isinstance(ndarray, np.ndarray):
        raise TypeError(name + ' must be an instance of ndarray')
    if ndarray.ndim != 1:
        raise ValueError(name + ' must be 1D')
    if size is not None:
        if ndarray.size != size:
            raise ValueError(name + ' must have size ' + str(size))
    if norm:
        if np.any(ndarray < 0) or np.any(ndarray > 1):
            raise ValueError(name + ' must be in the interval [0,1]')
    if positive:
        if np.any(ndarray < 0):
            raise ValueError(name + ' must be positive')
    if limit is not None:
        if np.any(ndarray < limit[0]):
            raise ValueError(name + ' must be greater or equal than ' + str(limit[0]))
        if np.any(ndarray > limit[1]):
            raise ValueError(name + ' must be smaller or equal than ' + str(limit[1]))
    if dtype is not None:
        if ndarray.dtype != dtype:
            raise TypeError(name + ' dtype must be ' + str(dtype))

def _validate_ndarray_2D(name, ndarray, shape0=None, shape1=None, norm=False, positive=False, limit=None, dtype=None):

    if not isinstance(ndarray, np.ndarray):
        raise TypeError(name + ' must be an instance of ndarray')
    if ndarray.ndim != 2:
        raise ValueError(name + ' must be 2D')
    if shape0 is not None:
        if ndarray.shape[0] != shape0:
            raise ValueError(name + ' must have dimension 0=' + str(shape0))
    if shape1 is not None:
        if ndarray.shape[1] != shape1:
            raise ValueError(name + ' must have dimension 1=' + str(shape1))
    if norm:
        if np.any(ndarray < 0) or np.any(ndarray > 1):
            raise ValueError(name + ' must be in the interval [0,1]')
    if positive:
        if np.any(ndarray < 0):
            raise ValueError(name + ' must be positive')
    if limit is not None:
        if np.any(ndarray < limit[0]):
            raise ValueError(name + ' must be greater or equal than ' + str(limit[0]))
        if np.any(ndarray > limit[1]):
            raise ValueError(name + ' must be smaller or equal than ' + str(limit[1]))
    if dtype is not None:
        if ndarray.dtype != dtype:
            raise TypeError(name + ' dtype must be ' + str(dtype))

def _validate_ndarray_3D(name, ndarray, shape0=None, shape1=None, shape2=None, norm=False, positive=False, limit=None, dtype=None):

    if not isinstance(ndarray, np.ndarray):
        raise TypeError(name + ' must be an instance of ndarray')
    if ndarray.ndim != 3:
        raise ValueError(name + ' must be 3D')
    if shape0 is not None:
        if ndarray.shape[0] != shape0:
            raise ValueError(name + ' must have dimension 0=' + str(shape0))
    if shape1 is not None:
        if ndarray.shape[1] != shape1:
            raise ValueError(name + ' must have dimension 1=' + str(shape1))
    if shape2 is not None:
        if ndarray.shape[2] != shape2:
            raise ValueError(name + ' must have dimension 2=' + str(shape1))
    if norm:
        if np.any(ndarray < 0) or np.any(ndarray > 1):
            raise ValueError(name + ' must be in the interval [0,1]')
    if positive:
        if np.any(ndarray < 0):
            raise ValueError(name + ' must be positive')
    if limit is not None:
        if np.any(ndarray < limit[0]):
            raise ValueError(name + ' must be greater or equal than ' + str(limit[0]))
        if np.any(ndarray > limit[1]):
            raise ValueError(name + ' must be smaller or equal than ' + str(limit[1]))
    if dtype is not None:
        if ndarray.dtype != dtype:
            raise TypeError(name + ' dtype must be ' + str(dtype))

def _validate_ndarray_4D(name, ndarray, shape0=None, shape1=None, shape2=None, shape3=None, norm=False, positive=False, limit=None, dtype=None):

    if not isinstance(ndarray, np.ndarray):
        raise TypeError(name + ' must be an instance of ndarray')
    if ndarray.ndim != 4:
        raise ValueError(name + ' must be 4D')
    if shape0 is not None:
        if ndarray.shape[0] != shape0:
            raise ValueError(name + ' must have dimension 0=' + str(shape0))
    if shape1 is not None:
        if ndarray.shape[1] != shape1:
            raise ValueError(name + ' must have dimension 1=' + str(shape1))
    if shape2 is not None:
        if ndarray.shape[2] != shape2:
            raise ValueError(name + ' must have dimension 2=' + str(shape1))
    if shape3 is not None:
        if ndarray.shape[3] != shape2:
            raise ValueError(name + ' must have dimension 3=' + str(shape1))
    if norm:
        if np.any(ndarray < 0) or np.any(ndarray > 1):
            raise ValueError(name + ' must be in the interval [0,1]')
    if positive:
        if np.any(ndarray < 0):
            raise ValueError(name + ' must be positive')
    if limit is not None:
        if np.any(ndarray < limit[0]):
            raise ValueError(name + ' must be greater or equal than ' + str(limit[0]))
        if np.any(ndarray > limit[1]):
            raise ValueError(name + ' must be smaller or equal than ' + str(limit[1]))
    if dtype is not None:
        if ndarray.dtype != dtype:
            raise TypeError(name + ' dtype must be ' + str(dtype))

def _validate_string(name, string, choices=None):

    if not isinstance(string, str):
        raise TypeError(name + ' must be an instance of str')
    if choices is not None and string not in choices:
        raise ValueError(name + ' must be one of the following: ' + str(choices))


def _validate_echogram(echogram):
    from masp.utils import C

    if not isinstance(echogram, Echogram):
        raise TypeError('echogram must be an instance of Echogram')

    _validate_ndarray_2D('echogram.value', echogram.value)
    _validate_ndarray_1D('echogram.time', echogram.time, positive=True)
    _validate_ndarray_2D('echogram.order', echogram.order, shape1=C, dtype=int)
    _validate_ndarray_2D('echogram.coords', echogram.coords, shape1=C)

    shapes = [echogram.value.shape[0], echogram.time.shape[0], echogram.order.shape[0],  echogram.coords.shape[0]]
    if not all(s == shapes[0] for s in shapes):
        raise ValueError('echogram shape mismatch')


def _validate_quantised_echogram(qechogram):
    from masp.utils import C

    if not isinstance(qechogram, QuantisedEchogram):
        raise TypeError('quantised echogram must be an instance of QuantisedEchogram')

    _validate_ndarray_2D('qechogram.value', qechogram.value)
    _validate_ndarray_1D('qechogram.time', qechogram.time, positive=True)
    _validate_boolean('qechogram.isActive', qechogram.isActive)

    if qechogram.value.shape[0] != qechogram.time.shape[0]:
        raise ValueError('quantised echogram shape mismatch')

def _validate_echogram_array(ndarray, shape0=None, shape1=None, shape2=None):
    """
    specific case of 2D/3D ndarray with dtype=Echogram
    """
    if not isinstance(ndarray, np.ndarray):
        raise TypeError('Echogram array must be an instance of ndarray')
    if ndarray.ndim not in (2, 3):
        raise ValueError('Echogram array must be 2D or 3D')
    if shape0 is not None:
        if ndarray.shape[0] != shape0:
            raise ValueError('Echogram array must have dimension 0=' + str(shape0))
    if shape1 is not None:
        if ndarray.shape[1] != shape1:
            raise ValueError('Echogram array must have dimension 1=' + str(shape1))
    if shape2 is not None:
        if ndarray.shape[2] != shape1:
            raise ValueError('Echogram array must have dimension 2=' + str(shape2))
    if ndarray.dtype != Echogram or not isinstance(ndarray.flatten()[0], Echogram):
        raise TypeError('Echogram array dtype must be Echogram')
    for idx in np.ndindex(ndarray.shape):
        _validate_echogram(ndarray[idx])


def _validate_quantised_echogram_array(ndarray, size=None):
    """
    specific case of 1D ndarray with dtype=QuantisedEchogram
    """
    if not isinstance(ndarray, np.ndarray):
        raise TypeError('Quantised Echogram array must be an instance of ndarray')
    if ndarray.ndim != 1:
        raise ValueError('Quantised Echogram array must be 1D')
    if size is not None:
        if ndarray.size != size:
            raise ValueError('Quantised Echogram array must have size=' + str(size))
    if ndarray.dtype != QuantisedEchogram  or not isinstance(ndarray.flatten()[0], QuantisedEchogram):
        raise TypeError('Quantised Echogram array dtype must be QuantisedEchogram')
    for idx in np.ndindex(ndarray.shape):
        _validate_quantised_echogram(ndarray[idx])



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
# Copyright (c) 2019, Eurecat / UPF
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the <organization> nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
#   @file   evaluate_sht_filters.py
#   @author Andrés Pérez-López
#   @date   01/10/2019
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

import numpy as np
import matplotlib.pyplot as plt
from masp.validate_data_types import _validate_int, _validate_ndarray_1D, \
    _validate_ndarray_2D, _validate_ndarray_3D, _validate_boolean, _validate_number


def evaluate_sht_filters(M_mic2sh, H_array, fs, Y_grid, w_grid=None, plot=False):
    """
    Evaluate frequency-dependent performance of SHT filters.

    Parameters
    ----------
    M_mic2sh : ndarray
        SHT filtering matrix produced by one of the methods included in the library.
        Dimension = ( (order+1)^2, nMics, nBins ).
    H_array : ndarray, dtype = 'complex'
         Modeled or measured spherical array responses in a dense grid of `nGrid` directions.
        Dimension = ( nBins, nMics, nGrid ).
    fs : int
        Target sampling rate.
    Y_grid : ndarray
        Spherical harmonics matrix for the `nGrid` directions of the evaluation grid.
        Dimension = ( nGrid, (order+1)^2 ).
    w_grid : ndarray, optional
        Vector of integration weights for the grid points.
        Dimension = ( nGrid ).
    plot : bool, optional
        Plot responses. Default to false.

    Returns
    -------
    cSH : ndarray, dtype = 'complex'
        Spatial correlation coefficient, for each SHT order and frequency bin.
        Dimension = ( nBins, order+1 ).
    lSH : ndarray
        Level difference, for each SHT order, for each SHT order and frequency bin.
        Dimension = ( nBins, order+1 ).
    WNG : ndarray
        Maximum amplification of all output SH components.
        Dimension = ( nBins ).

    Raises
    -----
    TypeError, ValueError: if method arguments mismatch in type, dimension or value.

    Notes
    -----
    The SHT filters can be evaluated in terms of how ideal are the SH
    components that they generate. The evaluation here follows the metrics
    introduced in

        Moreau, S., Daniel, J., Bertet, S., 2006,
        `3D sound field recording with higher order ambisonics-objectiv
        measurements and validation of spherical microphone.`
        In Audio Engineering Society Convention 120.

    These are a) the spatial correlation coefficient between each ideal
    spherical harmonic and the reconstructed pattern, evaluated at a dense
    grid of directions, b) level difference between the mean spatial power
    of the reconstructed pattern (diffuse power) over the one from an ideal
    SH component. Ideally, correlaiton should be close to one, and the
    level difference should be close to 0dB.

    Additionally, the maximum amplification of all output SH components is
    evaluated, through the maximum eigenvalue of the filtering matrix.

    Due to the matrix nature of computations,
    the minimum valid value for `nMics` and `nGrid` is 2.
    """

    _validate_ndarray_3D('M_mic2sh', M_mic2sh)
    n_sh = M_mic2sh.shape[0]
    order_sht = int(np.sqrt(n_sh) - 1)
    nMics = M_mic2sh.shape[1]
    _validate_number('nMics', nMics, limit=[2, np.inf])
    nBins = M_mic2sh.shape[2]

    _validate_ndarray_3D('H_array', H_array, shape0=nBins, shape1=nMics)
    nGrid = H_array.shape[2]
    _validate_number('nGrid', nGrid, limit=[2, np.inf])

    _validate_ndarray_2D('Y_grid', Y_grid, shape0=nGrid, shape1=n_sh)

    if w_grid is None:
        w_grid = 1/nGrid*np.ones(nGrid)
    _validate_ndarray_1D('w_grid', w_grid, size=nGrid)

    _validate_int('fs', fs, positive=True)
    if plot is not None:
        _validate_boolean('plot', plot)

    nFFT = 2 * (nBins - 1)
    f = np.arange(nFFT // 2 + 1) * fs / nFFT

    # Compute spatial correlations and integrated level difference between
    # ideal and reconstructed harmonics
    cSH = np.empty((nBins, order_sht+1), dtype='complex')
    lSH = np.empty((nBins, order_sht+1))
    # rSH = np.empty((nBins, order_sht+1))
    for kk in range(nBins):
        H_kk = H_array[kk,:,:]
        y_recon_kk = np.matmul(M_mic2sh[:,:, kk], H_kk)
        for n in range(order_sht+1):
            cSH_n = 0  # spatial correlation (mean per order)
            lSH_n = 0  # diffuse level difference (mean per order)
            # rSH_n = 0  # mean level difference (mean per order)
            for m in range(-n, n+1):
                q = np.power(n, 2) + n + m
                y_recon_nm = y_recon_kk[q,:].T
                y_ideal_nm = Y_grid[:, q]
                cSH_nm = np.matmul((y_recon_nm * w_grid).conj(), y_ideal_nm) / np.sqrt( np.matmul((y_recon_nm*w_grid).conj(), y_recon_nm ))
                cSH_n = cSH_n + cSH_nm
                lSH_nm = np.real(np.matmul((y_recon_nm * w_grid).conj(), y_recon_nm ))
                lSH_n = lSH_n + lSH_nm
                # rSH_nm = np.sum(np.power(np.abs(y_recon_nm - y_ideal_nm), 2) * w_grid)
                # rSH_n = rSH_n + rSH_nm;
            cSH[kk, n] = cSH_n / (2 * n + 1)
            lSH[kk, n] = lSH_n / (2 * n + 1)
            # rSH[kk, n] = rSH_n / (2 * n + 1)

    # Maximum noise amplification of all filters in matrix
    WNG = np.empty(nBins)
    for kk in range(nBins):
        # TODO: Matlab implementation warns when M matrix is complex, e.g. TEST_SCRIPTS l. 191-199
        # Avoid ComplexWarning: imaginary parts appearing due to numerical precission
        eigM = np.real(np.linalg.eigvals(np.matmul(M_mic2sh[:,:,kk].T.conj(), M_mic2sh[:,:,kk])))
        WNG[kk] = np.max(eigM)

    # Plots
    if plot:
        str_legend = [None]*(order_sht+1)
        for n in range(order_sht+1):
            str_legend[n] = str(n)

        plt.figure()
        plt.subplot(311)
        plt.semilogx(f, np.abs(cSH))
        plt.grid()
        plt.legend(str_legend)
        plt.axis([50, 20000, 0, 1])
        plt.title('Spatial correlation')

        plt.subplot(312)
        plt.semilogx(f, 10 * np.log10(lSH))
        plt.grid()
        plt.legend(str_legend)
        plt.axis([50, 20000, -30, 10])
        plt.title('Level correlation')

        plt.subplot(313)
        plt.semilogx(f, 10 * np.log10(WNG))
        plt.grid()
        plt.xlim([50, 20000])
        plt.title('Maximum amplification')
        plt.xlabel('Frequency (Hz)')

        # plt.subplot(414)
        # plt.semilogx(f, 10 * np.log10(rSH))
        # plt.grid()
        # plt.xlim([50, 20000])
        # plt.title('MSE')
        # plt.xlabel('Frequency (Hz)')

        plt.show()

    return cSH, lSH, WNG



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
# Copyright (c) 2019, Eurecat / UPF
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the <organization> nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
#   @file   image_source_method.py
#   @author Andrés Pérez-López
#   @date   30/07/2019
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

import numpy as np

from masp.validate_data_types import _validate_ndarray_1D, _validate_int, _validate_number, \
    _validate_echogram, _validate_string
from .echogram import Echogram
from masp.utils import C, c


def ims_coreMtx(room, source, receiver, type, typeValue):
    """
    Compute echogram by image source method.

    Parameters
    ----------
    room : ndarray
        Room dimensions in cartesian coordinates. Dimension = (3) [x, y, z].
    src : ndarray
        Source position in cartesian coordinates. Dimension = (3) [x, y, z].
    rec : ndarray
        Receiver position in cartesian coordinates. Dimension = (3) [x, y, z].
    type : str
        Restriction type: 'maxTime' or 'maxOrder'
    typeValue: int or float
        Value of the chosen restriction.

    Returns
    -------
    reflections : echogram
        An Echogram instance.

    Raises
    -----
    TypeError, ValueError: if method arguments mismatch in type, dimension or value.

    Notes
    -----
    `src` and `rec` positions are specified from the left ground corner
    of the room, using a left-handed coordinate system.
    `room` refers to the wall dimensions.
    Therefore, their values should be positive and smaller than room dimensions.

              _____    _
             |     |   |
             |     |   |
           x ^     |   | l = r[0]
             |     |   |
             |     |   |
             o---->    -
                  y
             |-----|
                w = r[1]

    """

    _validate_ndarray_1D('room', room, size=C, positive=True)
    _validate_ndarray_1D('source', source, size=C, positive=True, limit=[np.zeros(C),room])
    _validate_ndarray_1D('receiver', receiver, size=C, positive=True, limit=[np.zeros(C),room])
    _validate_string('type', type, choices=['maxTime', 'maxOrder'])

    # Room dimensions
    l, w, h = room

    # Move source origin to the centrer of the room
    src = np.empty(C)
    src[0] = source[0] - l / 2
    src[1] = w / 2 - source[1]
    src[2] = source[2] - h / 2

    # Move receiver origin to the centrer of the room
    rec = np.empty(C)
    rec[0] = receiver[0] - l / 2
    rec[1] = w / 2 - receiver[1]
    rec[2] = receiver[2] - h / 2

    if type is 'maxOrder':
        maxOrder = typeValue
        echogram = ims_coreN(room, src, rec, maxOrder)
    elif type is 'maxTime':
        maxDelay = typeValue
        echogram = ims_coreT(room, src, rec, maxDelay)

    # Sort reflections according to propagation time
    idx = np.argsort(echogram.time)
    echogram.time = echogram.time[idx]
    echogram.value = echogram.value[idx]
    echogram.order = echogram.order[idx, :]
    echogram.coords = echogram.coords[idx, :]

    return echogram


def ims_coreN(room, src, rec, N):
    """
    Compute echogram by image source method, under reflection order restriction

    Parameters
    ----------
    room : ndarray
        Room dimensions in cartesian coordinates. Dimension = (3) [x, y, z].
    src : ndarray
        Source position in cartesian coordinates. Dimension = (3) [x, y, z].
    rec : ndarray
        Receiver position in cartesian coordinates. Dimension = (3) [x, y, z].
    N : int
        Maximum reflection order.

    Returns
    -------
    reflections : echogram
        An Echogram instance.

    Raises
    -----
    TypeError, ValueError: if method arguments mismatch in type, dimension or value.

    Notes
    -----
    `src` and `rec` positions are specified from a right-handed coordinate system
     placed at the center of the room, with +x facing front, and +y facing left.
     (as opposite to `ims_coreMtx`).
     However, `room` refer to the wall dimensions.
     Therefore, given values must be in the range +-room[i]/2.

                ^x
              __|__    _
             |  |  |   |
             |  |  |   |
          y<----o  |   | l = r[0]
             |     |   |
             |     |   |
             |_____|   -

             |-----|
                w = r[1]

    """

    _validate_ndarray_1D('room', room, size=C, positive=True)
    _validate_ndarray_1D('source', src, size=C, limit=[-room/2,room/2])
    _validate_ndarray_1D('receiver', rec, size=C, limit=[-room/2,room/2])
    _validate_int('N', N, positive=True)

    # i,j,k indices for calculation in x,y,z respectively
    r = np.arange(-N, N+1)
    xx, yy, zz = np.meshgrid(r, r, r)
    # Vectorize (kind of empirical...)
    i = zz.reshape(zz.size)
    j = xx.reshape(xx.size)
    k = yy.reshape(yy.size)
    # Compute total order and select only valid incides up to order N
    s_ord = np.abs(i) + np.abs(j) + np.abs(k)
    i = i[s_ord <= N]
    j = j[s_ord <= N]
    k = k[s_ord <= N]

    # Image source coordinates with respect to receiver
    s_x = i*room[0] + np.power(-1.,i)*src[0] - rec[0]
    s_y = j*room[1] + np.power(-1.,j)*src[1] - rec[1]
    s_z = k*room[2] + np.power(-1.,k)*src[2] - rec[2]

    # Distance
    s_d = np.sqrt(np.power(s_x,2) + np.power(s_y,2) + np.power(s_z,2))
    # Reflection propagation time
    s_t = s_d/c
    # Reflection propagation attenuation - if distance is <1m
    # set at attenuation at 1 to avoid amplification
    s_att = np.empty(s_d.size)
    s_att[s_d <= 1] = 1
    s_att[s_d > 1] = 1./s_d[s_d > 1]

    # Write to echogram structure
    reflections = Echogram(value=s_att[:, np.newaxis],
                           time=s_t,
                           order=np.asarray(np.stack([i, j, k], axis=1), dtype=int),
                           coords=np.stack([s_x, s_y, s_z], axis=1))

    return reflections


def ims_coreT(room, src, rec, maxTime):
    """
    Compute echogram by image source method, under maxTime restriction

    Parameters
    ----------
    room : ndarray
        Room dimensions in cartesian coordinates. Dimension = (3) [x, y, z].
    src : ndarray
        Source position in cartesian coordinates. Dimension = (3) [x, y, z].
    rec : ndarray
        Receiver position in cartesian coordinates. Dimension = (3) [x, y, z].
    maxTime : float
        Maximum echogram computation time.

    Returns
    -------
    reflections : echogram
        An Echogram instance.

    Raises
    -----
    TypeError, ValueError: if method arguments mismatch in type, dimension or value.

    Notes
    -----
    `src` and `rec` positions are specified from a right-handed coordinate system
     placed at the center of the room, with +x facing front, and +y facing left.
     (as opposite to `ims_coreMtx`).
     However, `room` refer to the wall dimensions.
     Therefore, given values must be in the range +-room[i]/2.

                ^x
              __|__    _
             |  |  |   |
             |  |  |   |
          y<----o  |   | l = r[0]
             |     |   |
             |     |   |
             |_____|   -

             |-----|
                w = r[1]

    """

    _validate_ndarray_1D('room', room, size=C, positive=True)
    _validate_ndarray_1D('source', src, size=C, limit=[-room/2,room/2])
    _validate_ndarray_1D('receiver', rec, size=C, limit=[-room/2,room/2])
    _validate_number('maxTime', maxTime, positive=True)

    # Find order N that corresponds to maximum distance
    d_max = maxTime * c
    Nx = np.ceil(d_max / room[0])
    Ny = np.ceil(d_max / room[1])
    Nz = np.ceil(d_max / room[2])

    # i, j, k indices for calculation in x, y, z respectively
    rx = np.arange(-Nx, Nx + 1)
    ry = np.arange(-Ny, Ny + 1)
    rz = np.arange(-Nz, Nz + 1)
    xx, yy, zz = np.meshgrid(rx, ry, rz)
    # Vectorize (transpose idx due to matlab/python variations on matrix handling)
    i = xx.transpose(2,0,1).flatten()
    j = yy.transpose(2,0,1).flatten()
    k = zz.transpose(2,0,1).flatten()
    # Image source coordinates with respect to receiver
    s_x = i*room[0] + np.power(-1.,i)*src[0] - rec[0]
    s_y = j*room[1] + np.power(-1.,j)*src[1] - rec[1]
    s_z = k*room[2] + np.power(-1.,k)*src[2] - rec[2]
    # Distance
    s_d = np.sqrt(np.power(s_x,2) + np.power(s_y,2) + np.power(s_z,2))

    # Bypass image sources with d > dmax
    i = i[s_d < d_max]
    j = j[s_d < d_max]
    k = k[s_d < d_max]
    s_x = s_x[s_d < d_max]
    s_y = s_y[s_d < d_max]
    s_z = s_z[s_d < d_max]
    s_d = s_d[s_d < d_max]

    # Reflection propagation time
    s_t = s_d/c
    # Reflection propagation attenuation - if distance is <1m
    # set at attenuation at 1 to avoid amplification
    s_att = np.zeros(s_d.size)
    s_att[s_d <= 1] = 1
    s_att[s_d > 1] = 1./s_d[s_d > 1]

    # Write to echogram structure
    reflections = Echogram(value=s_att[:, np.newaxis],
                           time=s_t,
                           order=np.asarray(np.stack([i, j, k], axis=1), dtype=int),
                           coords=np.stack([s_x, s_y, s_z], axis=1))

    return reflections



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
# Copyright (c) 2019, Eurecat / UPF
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the <organization> nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
#   @file   array_sht_filters.py
#   @author Andrés Pérez-López
#   @date   02/10/2019
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

import numpy as np
import warnings
from masp.utils import c, C, elev2incl, get_sh, replicate_per_order
from masp.array_response_simulator import sph_modal_coefs
from masp.validate_data_types import _validate_int, _validate_float, _validate_ndarray_2D, _validate_ndarray_1D, \
    _validate_ndarray_3D, _validate_number


def array_sht_filters_theory_radInverse(R, nMic, order_sht, Lfilt, fs, amp_threshold):
    """
    Generate SHT filters based on theoretical responses (regularized radial inversion)

    Parameters
    ----------
    R : float
        Microphone array radius, in meter.
    nMic : int
        Number of microphone capsules.
    order_sht : int
        Spherical harmonic transform order.
    Lfilt : int
        Number of FFT points for the output filters. It must be even.
    fs : int
        Sample rate for the output filters.
    amp_threshold : float
         Max allowed amplification for filters, in dB.

    Returns
    -------
    h_filt : ndarray
        Impulse responses of the filters. Dimension = (Lfilt, order_sht+1).
    H_filt: ndarray
        Frequency-domain filters. Dimension = (Lfilt//2+1, order_sht+1).

    Raises
    -----
    TypeError, ValueError: if method arguments mismatch in type, dimension or value.
    UserWarning: if `nMic` not big enough for the required sht order.

    Notes
    -----
    Generate the filters to convert microphone signals from a spherical
    microphone array to SH signals, based on an ideal theoretical model of
    the array. The filters are generated from an inversion of the radial
    components of the response, neglecting spatial aliasing effects and
    non-ideal arrangements of the microphones. One filter is shared by all
    SH signals of the same order in this case.
    Here this single channel inversion problem is done with a constraint on
    max allowed amplification regularization, using Tikhonov
    regularization, similar e.g. to:

        Moreau, S., Daniel, J., Bertet, S., 2006,
        3D sound field recording with higher order ambisonics-objective
        measurements and validation of spherical microphone.
        In Audio Engineering Society Convention 120.

    """

    _validate_float('R', R, positive=True)
    _validate_int('nMic', nMic, positive=True)
    _validate_int('order_sht', order_sht, positive=True)
    _validate_int('Lfilt', Lfilt, positive=True, parity='even')
    _validate_int('fs', fs, positive=True)
    _validate_float('amp_threshold', amp_threshold)

    f = np.arange(Lfilt // 2 + 1) * fs / Lfilt

    # Adequate sht order to the number of microphones
    if order_sht > np.sqrt(nMic) - 1:
        order_sht = int(np.floor(np.sqrt(nMic) - 1))
        warnings.warn("Set order too high for the number of microphones, should be N<=np.sqrt(Q)-1. Auto set to "+str(order_sht), UserWarning)

    # Modal responses
    kR = 2 * np.pi * f * R / c
    bN = sph_modal_coefs(order_sht, kR, 'rigid') / (4 * np.pi)

    # Encoding matrix with regularization
    a_dB = amp_threshold
    alpha = complex(np.sqrt(nMic) * np.power(10, a_dB / 20))  # Explicit casting to allow negative sqrt (a_dB < 0)
    beta = np.sqrt((1 - np.sqrt(1 - 1 / np.power(alpha,2))) / (1 + np.sqrt(1 - 1 / np.power(alpha,2))))  # Moreau & Daniel
    # Regularized single channel equalization filters per order
    H_filt = (bN).conj() / (np.power(np.abs(bN), 2) + np.power(beta, 2) * np.ones((Lfilt // 2 + 1, order_sht + 1)))

    # Time domain filters
    temp = H_filt.copy()
    temp[-1,:] = np.real(temp[-1,:])
    # Create the symmetric conjugate negative frequency response for a real time-domain signal
    h_filt = np.real(np.fft.fftshift(np.fft.ifft(np.append(temp, np.conj(temp[-2:0:-1,:]), axis=0), axis=0), axes=0))

    # TODO: check return ordering
    return h_filt, H_filt


def array_sht_filters_theory_softLim(R, nMic, order_sht, Lfilt, fs, amp_threshold):
    """
    Generate SHT filters based on theoretical responses (soft-limiting)

    Parameters
    ----------
    R : float
        Microphone array radius, in meter.
    nMic : int
        Number of microphone capsules.
    order_sht : int
        Spherical harmonic transform order.
    Lfilt : int
        Number of FFT points for the output filters. It must be even.
    fs : int
        Sample rate for the output filters.
    amp_threshold : float
         Max allowed amplification for filters, in dB.

    Returns
    -------
    h_filt : ndarray
        Impulse responses of the filters. Dimension = (Lfilt, order_sht+1).
    H_filt: ndarray
        Frequency-domain filters. Dimension = (Lfilt//2+1, order_sht+1).

    Raises
    -----
    TypeError, ValueError: if method arguments mismatch in type, dimension or value.
    UserWarning: if `nMic` not big enough for the required sht order.

    Notes
    -----
    Generate the filters to convert microphone signals from a spherical
    microphone array to SH signals, based on an ideal theoretical model of
    the array. The filters are generated from an inversion of the radial
    components of the response, neglecting spatial aliasing effects and
    non-ideal arrangements of the microphones. One filter is shared by all
    SH signals of the same order in this case.
    Here this single channel inversion problem is done through a
    thresholding approach on the inversion, limited to a max allowed
    amplification. The limiting follows the approach of

        Bernschutz, B., Porschmann, C., Spors, S., Weinzierl, S., Versterkung, B., 2011.
        Soft-limiting der modalen amplitudenverst?rkung bei sph?rischen mikrofonarrays im plane wave decomposition verfahren.
        Proceedings of the 37. Deutsche Jahrestagung fur Akustik (DAGA 2011)

    """

    _validate_float('R', R, positive=True)
    _validate_int('nMic', nMic, positive=True)
    _validate_int('order_sht', order_sht, positive=True)
    _validate_int('Lfilt', Lfilt, positive=True, parity='even')
    _validate_int('fs', fs, positive=True)
    _validate_float('amp_threshold', amp_threshold)

    f = np.arange(Lfilt // 2 + 1) * fs / Lfilt

    # Adequate sht order to the number of microphones
    if order_sht > np.sqrt(nMic) - 1:
        order_sht = int(np.floor(np.sqrt(nMic) - 1))
        warnings.warn("Set order too high for the number of microphones, should be N<=np.sqrt(Q)-1. Auto set to "+str(order_sht), UserWarning)

    # Modal responses
    kR = 2 * np.pi * f * R / c
    bN = sph_modal_coefs(order_sht, kR, 'rigid') / (4 * np.pi)  # due to modified SHs, the 4pi term disappears from the plane wave expansion
    # Single channel equalization filters per order
    inv_bN = 1. / bN
    inv_bN[0, 1:] = 0.

    # Encoding matrix with regularization
    a_dB = amp_threshold
    alpha = complex(np.sqrt(nMic) * np.power(10, a_dB / 20))  # Explicit casting to allow negative sqrt (a_dB < 0)
    # Regularized single channel equalization filters per order
    H_filt = (2*alpha/np.pi) * (np.abs(bN) * inv_bN) * np.arctan((np.pi / (2 * alpha))* np.abs(inv_bN))

    # Time domain filters
    temp = H_filt.copy()
    temp[-1,:] = np.real(temp[-1,:])
    # Create the symmetric conjugate negative frequency response for a real time-domain signal
    h_filt = np.real(np.fft.fftshift(np.fft.ifft(np.append(temp, np.conj(temp[-2:0:-1,:]), axis=0), axis=0), axes=0))

    # TODO: check return order
    return h_filt, H_filt


def array_sht_filters_theory_regLS(R, mic_dirsAziElev, order_sht, Lfilt, fs, amp_threshold):
    """
    Generate SHT filters based on theoretical responses (regularized least-squares)

    Parameters
    ----------
    R : float
        Microphone array radius, in meter.
    mic_dirsAziElev : ndarray
        Evaluation directions. Dimension = (nDirs, 2).
        Directions are expected in radians, expressed in pairs [azimuth, elevation].
    order_sht : int
        Spherical harmonic transform order.
    Lfilt : int
        Number of FFT points for the output filters. It must be even.
    fs : int
        Sample rate for the output filters.
    amp_threshold : float
         Max allowed amplification for filters, in dB.

    Returns
    -------
    h_filt : ndarray
        Impulse responses of the filters. Dimension = ( nMic, (order_sht+1)^2, Lfilt ).
    H_filt: ndarray
        Frequency-domain filters. Dimension = ( nMic, (order_sht+1)^2, Lfilt//2+1 ).

    Raises
    -----
    TypeError, ValueError: if method arguments mismatch in type, dimension or value.
    UserWarning: if `nMic` not big enough for the required sht order.
    TODO: ValueError: if the array order is not big enough.

    Notes
    -----
    Generate the filters to convert microphone signals from a spherical
    microphone array to SH signals, based on an ideal theoretical model of 
    the array. The filters are generated as a least-squares
    solution with a constraint on filter amplification, using Tikhonov
    regularization. The method formulates the LS problem in the spherical
    harmonic domain, by expressing the array response to an order-limited
    series of SH coefficients, similar to

        Jin, C.T., Epain, N. and Parthy, A., 2014.
        Design, optimization and evaluation of a dual-radius spherical microphone array.
        IEEE/ACM Transactions on Audio, Speech, and Language Processing, 22(1), pp.193-204.

    """

    _validate_float('R', R, positive=True)
    _validate_ndarray_2D('mic_dirsAziElev', mic_dirsAziElev, shape1=C-1)
    _validate_int('order_sht', order_sht, positive=True)
    _validate_int('Lfilt', Lfilt, positive=True, parity='even')
    _validate_int('fs', fs, positive=True)
    _validate_float('amp_threshold', amp_threshold)

    f = np.arange(Lfilt // 2 + 1) * fs / Lfilt
    num_f = f.size
    kR = 2 * np.pi * f * R / c
    kR_max = kR[-1]

    nMic = mic_dirsAziElev.shape[0]
    # Adequate sht order to the number of microphones
    if order_sht > np.sqrt(nMic) - 1:
        order_sht = int(np.floor(np.sqrt(nMic) - 1))
        warnings.warn("Set order too high for the number of microphones, should be N<=np.sqrt(Q)-1. Auto set to "+str(order_sht), UserWarning)

    mic_dirsAziIncl = elev2incl(mic_dirsAziElev)
    order_array = int(np.floor(2 * kR_max))

    # TODO: check validity of the approach
    if order_array <= 1:
        raise ValueError("Order array <= 1. Consider increasing R or fs.")

    Y_array = np.sqrt(4 * np.pi) * get_sh(order_array, mic_dirsAziIncl, 'real')

    # Modal responses
    bN = sph_modal_coefs(order_array, kR, 'rigid') / (4 * np.pi)  # due to modified SHs, the 4pi term disappears from the plane wave expansion

    # Array response in the SHD
    H_array = np.zeros((nMic, np.power(order_array+1, 2), num_f), dtype='complex')
    for kk in range(num_f):
        temp_b = bN[kk, :].T
        B = np.diag(replicate_per_order(temp_b))
        H_array[:, :, kk] = np.matmul(Y_array, B)

    a_dB = amp_threshold
    alpha = complex(np.power(10, a_dB / 20))  # Explicit casting to allow negative sqrt (a_dB < 0)
    beta = 1 / (2 * alpha)
    H_filt = np.zeros((np.power(order_sht+1,2), nMic, num_f), dtype='complex')
    for kk in range(num_f):
        tempH_N = H_array[:, :, kk]
        tempH_N_trunc = tempH_N[:, :np.power(order_sht+1,2)]
        H_filt[:, :, kk] = np.matmul(tempH_N_trunc.T.conj(), np.linalg.inv(np.matmul(tempH_N, tempH_N.T.conj()) + np.power(beta, 2) * np.eye(nMic)))

    # Time domain filters
    h_filt = H_filt.copy()
    h_filt[:, :, -1] = np.abs(h_filt[:, :, -1])
    h_filt = np.concatenate((h_filt, np.conj(h_filt[:, :, -2:0:-1])), axis=2)
    h_filt = np.real(np.fft.ifft(h_filt, axis=2))
    h_filt = np.fft.fftshift(h_filt, axes=2)

    # TODO: check return ordering
    return h_filt, H_filt


def array_sht_filters_measure_regLS(H_array, order_sht, grid_dirs_rad, w_grid=None, nFFT=1024, amp_threshold=10.):
    """
    Generate SHT filters based on measured responses (regularized least-squares)

    Parameters
    ----------
    H_array : ndarray, dtype=complex
        Frequency domain measured array responses. Dimension = ( nFFT//2+1, nMics, nGrids )
    order_sht : int
        Spherical harmonic transform order.
    grid_dirs_rad: ndarray,
        Grid positions in [azimuth, elevation] pairs (in radians). Dimension = (nGrid, 2).
    w_grid : ndarray, optional
        Weights for weighted-least square solution, based on the importance or area
        around each measurement point (leave empty if not known, or not important)
        Dimension = ( nGrid ).
    nFFT : int, optional
        Number of points for the FFT.
    amp_threshold : float
        Max allowed amplification for filters, in dB.

    Returns
    -------
    h_filt : ndarray
        Impulse responses of the filters. Dimension = ( (order_sht+1)^2, nMics, nFFT ).
    H_filt: ndarray
        Frequency-domain filters. Dimension = ( (order_sht+1)^2, nMics, nFFT//2+1 ).

    Raises
    -----
    TypeError, ValueError: if method arguments mismatch in type, dimension or value.
    UserWarning: if `nMic` not big enough for the required sht order.

    Notes
    -----
    Generate the filters to convert microphone signals from a spherical
    microphone array to SH signals, based on a least-squares solution with
    a constraint on noise amplification, using Tikhonov regularization. The
    method formulates the LS problem in the space domain, using the
    directional measurements of the array response, similar to e.g.

        Moreau, S., Daniel, J., Bertet, S., 2006,
        3D sound field recording with higher order ambisonics-objective
        measurements and validation of spherical microphone.
        In Audio Engineering Society Convention 120.

    Due to the matrix nature of computations,
    the minimum valid value for `nMic` and `nGrid` is 2.
    # TODO: nFFT argument is redundant!!!
    """

    _validate_int('nFFT', nFFT, positive=True, parity='even')
    nBins = nFFT // 2 + 1

    _validate_ndarray_3D('H_array', H_array, shape0=nBins)
    nMic = H_array.shape[1]
    _validate_number('nMic', nMic, limit=[2, np.inf])
    nGrid = H_array.shape[2]
    _validate_number('nGrid', nGrid, limit=[2, np.inf])

    _validate_int('order_sht', order_sht, positive=True)
    _validate_ndarray_2D('grid_dirs_rad', grid_dirs_rad, shape1=C-1)

    if w_grid is None:
        w_grid = 1/nGrid*np.ones(nGrid)
    _validate_ndarray_1D('w_grid', w_grid, size=nGrid)

    _validate_float('amp_threshold', amp_threshold)

    # Adequate sht order to the number of microphones
    if order_sht > np.sqrt(nMic) - 1:
        order_sht = int(np.floor(np.sqrt(nMic) - 1))
        warnings.warn(
            "Set order too high for the number of microphones, should be N<=np.sqrt(Q)-1. Auto set to " + str(
                order_sht), UserWarning)


    # SH matrix at grid directions
    Y_grid = np.sqrt(4 * np.pi) * get_sh(order_sht, elev2incl(grid_dirs_rad), 'real').T  # SH matrix for grid directions

    # Compute inverse matrix
    a_dB = amp_threshold
    alpha = complex(np.power(10, a_dB / 20))  # Explicit casting to allow negative sqrt (a_dB < 0)
    beta = 1 / (2 * alpha)
    W_grid = np.diag(w_grid)
    H_filt = np.zeros((np.power(order_sht+1,2), nMic, nBins), dtype='complex')
    for kk in range(nBins):
        tempH = H_array[kk, :, :]
        m = np.matmul(np.matmul(Y_grid, W_grid), tempH.T.conj())
        m2 = np.matmul(np.matmul(tempH, W_grid), tempH.T.conj())
        H_filt[:, :, kk] = np.matmul(m, np.linalg.inv(m2 + np.power(beta, 2) * np.eye(nMic)))

    # Time domain filters
    h_filt = H_filt.copy()
    h_filt[:, :, -1] = np.abs(h_filt[:, :, -1])
    h_filt = np.concatenate((h_filt, np.conj(h_filt[:, :, -2:0:-1])), axis=2)
    h_filt = np.real(np.fft.ifft(h_filt, axis=2))
    h_filt = np.fft.fftshift(h_filt, axes=2)

    # TODO: check return ordering
    return h_filt, H_filt


def array_sht_filters_measure_regLSHD(H_array, order_sht, grid_dirs_rad, w_grid=None, nFFT=1024, amp_threshold=10.):
    """
    Generate SHT filters based on measured responses  (regularized least-squares in the SHD)

    Parameters
    ----------
    H_array : ndarray, dtype=complex
        Frequency domain measured array responses. Dimension = ( nFFT//2+1, nMics, nGrids )
    order_sht : int
        Spherical harmonic transform order.
    grid_dirs_rad: ndarray,
        Grid positions in [azimuth, elevation] pairs (in radians). Dimension = (nGrid, 2).
    w_grid : ndarray, optional
        Weights for weighted-least square solution, based on the importance or area
        around each measurement point (leave empty if not known, or not important)
        Dimension = ( nGrid ).
    nFFT : int, optional
        Number of points for the FFT.
    amp_threshold : float
        Max allowed amplification for filters, in dB.

    Returns
    -------
    h_filt : ndarray
        Impulse responses of the filters. Dimension = ( (order_sht+1)^2, nMics, nFFT ).
    H_filt: ndarray
        Frequency-domain filters. Dimension = ( (order_sht+1)^2, nMics, nFFT//2+1 ).

    Raises
    -----
    TypeError, ValueError: if method arguments mismatch in type, dimension or value.
    UserWarning: if `nMic` not big enough for the required sht order.
    TODO: ValueError: if the array order is not big enough.

    Notes
    -----
    Generate the filters to convert micorphone signals from a spherical
    microphone array to SH signals, based on a least-squares solution with
    a constraint on noise amplification, using Tikhonov regularization. The
    method formulates the LS problem in the space domain, using the
    directional measurements of the array response, similar to e.g.

        Moreau, S., Daniel, J., Bertet, S., 2006,
        3D sound field recording with higher order ambisonics-objective
        measurements and validation of spherical microphone.
        In Audio Engineering Society Convention 120.

    # TODO: nFFT argument is redundant!!!
    Due to the matrix nature of computations,
    the minimum valid value for `nMics` and `nGrid` is 2 and 16 respectively.
    """

    _validate_int('nFFT', nFFT, positive=True, parity='even')
    nBins = nFFT // 2 + 1

    _validate_ndarray_3D('H_array', H_array, shape0=nBins)
    nMic = H_array.shape[1]
    _validate_number('nMic', nMic, limit=[2, np.inf])
    nGrid = H_array.shape[2]
    _validate_number('nGrid', nGrid, limit=[16, np.inf])

    _validate_int('order_sht', order_sht, positive=True)
    _validate_ndarray_2D('grid_dirs_rad', grid_dirs_rad, shape1=C-1)

    if w_grid is None:
        w_grid = 1/nGrid*np.ones(nGrid)
    _validate_ndarray_1D('w_grid', w_grid, size=nGrid)

    _validate_float('amp_threshold', amp_threshold)

    # Adequate sht order to the number of microphones
    if order_sht > np.sqrt(nMic) - 1:
        order_sht = int(np.floor(np.sqrt(nMic) - 1))
        warnings.warn(
            "Set order too high for the number of microphones, should be N<=np.sqrt(Q)-1. Auto set to " + str(
                order_sht), UserWarning)

    order_array = int(np.floor(np.sqrt(nGrid) / 2 - 1))
    # TODO: check validity of the approach
    # order_array must be greater or equal than requested order_sht ( nGrid > (2*(order_sht+1))^2 )
    if order_array < order_sht:
        raise ValueError("Order array < Order SHT. Consider increasing nGrid")

    # SH matrix at grid directions
    Y_grid = np.sqrt(4 * np.pi) * get_sh(order_array, elev2incl(grid_dirs_rad), 'real').T  # SH matrix for grid directions

    # Compute inverse matrix
    a_dB = amp_threshold
    alpha = complex(np.power(10, a_dB / 20))  # Explicit casting to allow negative sqrt (a_dB < 0)
    beta = 1 / (2 * alpha)
    W_grid = np.diag(w_grid)
    H_nm = np.zeros(( nBins, nMic, np.power(order_array+1,2) ), dtype='complex')
    for kk in range(nBins):
        tempH = H_array[kk, :, :]
        H_nm[kk, :, :] = np.matmul(np.matmul(np.matmul(tempH, W_grid), Y_grid.T),
                                   np.linalg.inv( np.matmul(np.matmul(Y_grid, W_grid), Y_grid.T)))

    # Compute the inverse matrix in the SHD with regularization
    H_filt = np.zeros((np.power(order_sht+1,2), nMic, nBins), dtype='complex')
    for kk in range(nBins):
        tempH_N = H_nm[kk, :, :]
        tempH_N_trunc = tempH_N[:, :np.power(order_sht+1, 2)]
        H_filt[:, :, kk] = np.matmul(tempH_N_trunc.T.conj(),
                                     np.linalg.inv(np.matmul(tempH_N, tempH_N.T.conj()) + np.power(beta, 2) * np.eye(nMic)))

    # Time domain filters
    h_filt = H_filt.copy()
    h_filt[:, :, -1] = np.abs(h_filt[:, :, -1])
    h_filt = np.concatenate((h_filt, np.conj(h_filt[:, :, -2:0:-1])), axis=2)
    h_filt = np.real(np.fft.ifft(h_filt, axis=2))
    h_filt = np.fft.fftshift(h_filt, axes=2)

    # TODO: check return ordering
    return h_filt, H_filt



from .plot_functions import plot_mic_array
from .sph_array_characteristics import sph_array_noise
from .sph_array_characteristics import sph_array_noise_threshold
from .sph_array_characteristics import sph_array_alias_lim
from .evaluate_sht_filters import evaluate_sht_filters
from .array_sht_filters import array_sht_filters_theory_radInverse
from .array_sht_filters import array_sht_filters_theory_softLim
from .array_sht_filters import array_sht_filters_theory_regLS
from .array_sht_filters import array_sht_filters_measure_regLS
from .array_sht_filters import array_sht_filters_measure_regLSHD
